#include <stdio.h>
#include <math.h>       /* for M_PI */
#include <unistd.h>     /* for usleep(2) */
#include "anim.h"

/* private funcitons */
static unsigned long AnimAllocColor(char *colName)
{
  Colormap cmap;
  XColor c0,c1;

  cmap=DefaultColormap(AnimStruct.display,0);
  XAllocNamedColor(AnimStruct.display,cmap,colName,&c1,&c0);
  return(c1.pixel);
}

static void AnimExposeScreen(void)
{
  XGCValues gcv;

  XCopyArea(AnimStruct.display,AnimStruct.pixmap,
            AnimStruct.window,AnimStruct.tmpgc,
            0,0,AnimStruct.width,AnimStruct.height,0,0);
}

static void AnimGraphicInit()
{
  Display *d;
  Window w;
  Pixmap p;

  d=AnimStruct.display=XOpenDisplay(NULL);
  w=AnimStruct.window=XCreateSimpleWindow(d,DefaultRootWindow(d),
                                          0,0,
                                          AnimStruct.width,AnimStruct.height,
                                          2,BlackPixel(d,0),WhitePixel(d,0));
  p=AnimStruct.pixmap=XCreatePixmap(d,DefaultRootWindow(d),
                                    AnimStruct.width,AnimStruct.height,
                                    DefaultDepth(d,0));
  AnimStruct.gc=XCreateGC(d,w,0,0);
  AnimStruct.tmpgc=XCreateGC(d,w,0,0);
  AnimStruct.normalColor=AnimAllocColor("red");
  AnimStruct.hilightColor=AnimAllocColor("blue");
  AnimStruct.backgroundColor=AnimAllocColor("white");
  /* ̤Υꥢ */
  XSetForeground(AnimStruct.display,AnimStruct.gc,AnimStruct.backgroundColor);
  XFillRectangle(d,p,AnimStruct.gc,0,0,AnimStruct.width,AnimStruct.height);
  XMapWindow(d,w);
  XSelectInput(d,w,ExposureMask);
  XFlush(d);
}

static int AnimPosX(double x)
{
  return((int)(AnimStruct.width*(x/(double)AnimStruct.size)));
}

static int AnimPosY(double y)
{
  return((int)(AnimStruct.height*((double)(AnimStruct.max-y))/AnimStruct.max));
}

static void AnimDrawDataX(int x,int y,int offset,unsigned long color)
{
  XEvent event;

  XSetForeground(AnimStruct.display,AnimStruct.gc,color);

  XFillRectangle(AnimStruct.display,AnimStruct.pixmap,AnimStruct.gc,
                 x,y+offset,
                 (int)(AnimStruct.width/(double)AnimStruct.size),
                 (int)(AnimStruct.height-y));

  XCopyArea(AnimStruct.display,AnimStruct.pixmap,
            AnimStruct.window,AnimStruct.tmpgc,
            x,0,
            (int)(AnimStruct.width/(double)AnimStruct.size),
            (int)(AnimStruct.height),
            x,0);
  XFlush(AnimStruct.display);
  XNextEvent(AnimStruct.display,&event);
  switch(event.type)
  {
   case Expose:
    AnimExposeScreen();
    break;
   defalut:
    break;
  }
}

static void AnimDrawData(int pos,unsigned long color)
{
  AnimDrawDataX(AnimPosX(pos),AnimPosY(AnimStruct.data[pos]),0,color);
}

/* public functions */
void AnimSetWaitTime(int time)
{
  AnimStruct.wait=(u_int)time;
}

void AnimSetView(int value)
{
  AnimStruct.view=value;
}

void AnimSetAnimation(int value)
{
  AnimStruct.anim=value;
}

void AnimInit(int dataSize,int view)
{
  AnimStruct.view=view;
  AnimStruct.anim=ANIM;
  AnimStruct.size=dataSize;
  AnimStruct.max=0;
  AnimStruct.width=ANIM_WIDTH;
  AnimStruct.height=ANIM_HEIGHT;
  AnimStruct.wait=(u_int)ANIM_WAIT_TIME;
  AnimStruct.swapCount=0;
  AnimStruct.hilightCount=0;
  /*
   AnimStruct.obj=(Object *)malloc(sizeof(Object)*dataSize);
   */
  AnimStruct.data=(int *)malloc(sizeof(int)*dataSize);
  if(AnimStruct.view)
    AnimGraphicInit(dataSize);
}

void AnimInitData(int pos,int value)
{
  if(AnimStruct.view)
  {
    AnimStruct.data[pos]=value;
    if(value>=AnimStruct.max)
      AnimStruct.max=value;
  }
}

void AnimShowData(int dummy)
{
  int i;
  
  if(AnimStruct.view)
    for(i=0;i<AnimStruct.size;i++)
      AnimDrawData(i,AnimStruct.normalColor);
  gettimeofday(&AnimStruct.before, (struct timezone *)NULL);
}


void AnimSwapData(int i,int j)
{
  int tmp;
  
  AnimStruct.swapCount++;
  if(AnimStruct.view)
  {
    AnimHilightData(i);
    AnimHilightData(j);
    usleep(AnimStruct.wait);

    if(AnimStruct.anim)
    {
      int k;
      int min=AnimMin(i,j);
      int max=AnimMax(i,j);
      double step=(max-min)/(double)(ANIM_STEP-1.0);
      XGCValues gcv;

      /* Erase Data for animation */
      AnimDrawData(i,AnimStruct.backgroundColor);
      AnimDrawData(j,AnimStruct.backgroundColor);

      gcv.function=GXxor;
      XChangeGC(AnimStruct.display,AnimStruct.gc,GCFunction,&gcv);

      for(k=0;k<ANIM_STEP;k++)
      {
        AnimDrawDataX(AnimPosX(min+k*step),AnimPosY(AnimStruct.data[min])
                      ,-(int)(ANIM_HOP*sin(M_PI*k/ANIM_STEP))
                      ,AnimStruct.hilightColor);
        AnimDrawDataX(AnimPosX(max-k*step),AnimPosY(AnimStruct.data[max])
                      ,-(int)(ANIM_HOP*sin(M_PI*k/ANIM_STEP))
                      ,AnimStruct.hilightColor);
        usleep(AnimStruct.wait/ANIM_STEP);
        AnimDrawDataX(AnimPosX(min+k*step),AnimPosY(AnimStruct.data[min])
                      ,-(int)(ANIM_HOP*sin(M_PI*k/ANIM_STEP))
                      ,AnimStruct.hilightColor);
        AnimDrawDataX(AnimPosX(max-k*step),AnimPosY(AnimStruct.data[max])
                      ,-(int)(ANIM_HOP*sin(M_PI*k/ANIM_STEP))
                      ,AnimStruct.hilightColor);
      }

      gcv.function=GXcopy;
      XChangeGC(AnimStruct.display,AnimStruct.gc,GCFunction,&gcv);
    }
    /* Erase Old Data */
    AnimDrawData(i,AnimStruct.backgroundColor);
    AnimDrawData(j,AnimStruct.backgroundColor);
    
    /* Swap data */
    tmp=AnimStruct.data[i];
    AnimStruct.data[i]=AnimStruct.data[j];
    AnimStruct.data[j]=tmp;
    
    /* Draw Swapped Data */
    AnimDrawData(i,AnimStruct.normalColor);
    AnimDrawData(j,AnimStruct.normalColor);
  }
}

void AnimHilightData(int pos)
{
  if(AnimStruct.view)
  {
    AnimDrawData(pos,AnimStruct.hilightColor);
    usleep(AnimStruct.wait);
  }
}

void AnimUnhilightData(int pos)
{
  AnimStruct.hilightCount++;
  if(AnimStruct.view)
  {
    AnimDrawData(pos,AnimStruct.normalColor);
  }
}

void AnimShowStats()
{
  struct timeval after;
  printf("Swap count:\t%d\n",AnimStruct.swapCount);
  printf("Hilight count:\t%d\n",AnimStruct.hilightCount);
  gettimeofday(&after, (struct timezone *)NULL);
  after.tv_sec -= AnimStruct.before.tv_sec;
  after.tv_usec -= AnimStruct.before.tv_usec;
  if (after.tv_usec < 0)
    after.tv_sec--, after.tv_usec += 1000000;
  printf("Sorting time:\t%ld.%04ld[sec]\n",after.tv_sec,after.tv_usec/100);
}
